# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause

"""Data for the unit tests for the feature-check Python library."""

FEATURES = { {{#each simple.features}} "{{@key}}": "{{this}}", {{/each}} }

SYNONYMS = {
    {{#each simple.synonyms}}
        "{{this.name}}": "{{this.sym}}",
    {{/each}}
}

COMPARE = [
    {{#each simple.compare}}
        ("{{this.name}}", "{{this.op}}", "{{this.version}}", {{pybool this.result}}),
    {{/each}}
]
